/*-
 * Copyright (C)2005 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: popushd.c,v 1.8 2007/07/28 19:25:20 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/event.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdarg.h>
#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include "popushd.h"

u_int32_t state;

void
mesg(u_int32_t flag, const char *fmt, const char *fmt1, ...) {
    va_list ap;

    if (flag & (POPUSH_CRI | BLIN_VER1)) {
        va_start(ap, fmt1);
        if  (flag & POPUSH_DMN) {
            if  (flag & POPUSH_ERX) vwarnx(fmt, ap);
            else vwarn(fmt, ap);
        } else {
            if  (flag & POPUSH_ERX) vsyslog((flag & POPUSH_CRI) ? LOG_CRIT : LOG_ERR, fmt, ap);
            else vsyslog((flag & POPUSH_CRI) ? LOG_CRIT : LOG_ERR, fmt1, ap);
        }
        va_end(ap);
}   }

static void
sighup(int sig) {
    state |= F_SIGN;
}

static void
sigint(int sig) {
    state |= F_STOP;
}

static void
usage(int ex) {
    fprintf(stderr, "popushd @BABOLO V.M " VERS "  " DATE "\n"
"Usage:\n"
"popushd [-qv ][-f conf][-p pid]\n"
"    -. envname - for cwd\n"
"    -@ envname - for PID\n"
"    -a 0xXX - absolute quote\n"
"    -d - not daeminize\n"
"    -f configfile\n"
"    -p PIDfile\n"
"    -q - quiet\n"
"    -v[v[v...]] - verbose\n"
);
    exit(ex);
}

int
main(int argc, char *argv[], char **envp) {
    int ex = EX_OK;
    int c;
    popush_main rr;
    popush_main *r = &rr;
    popush_cnf *e;
    struct kevent evnt;
    char *wdpath;
    pid_t mypid = getpid();
    char myspid[8];

    r->actionbit = NULL;
    bzero(r, sizeof(popush_main));
    r->flag = BLIN_VER1;
    r->cnfname = RCFILE;
    r->pidname = PIDFILE;
    while ((c = getopt(argc, argv, ".:@:a:df:p:qv")) != EOF) {
        switch (c) {
        case '.':
            wdpath = getcwd(NULL, 0);
            if  (!wdpath) {
                ifBLIN_QV1(r->flag) warn("getcwd failed");
            } else if (setenv(optarg, wdpath, 1) < 0) {
                ifBLIN_QV1(r->flag) warn("setenv %s=%s", optarg, wdpath);
            }
            break;
        case '@':
            if  (snprintf(myspid, 8, "%d", mypid) > 7) {
                ifBLIN_QV1(r->flag) warn("Too high pid %d", mypid);
            } else if (setenv(optarg, myspid, 1) < 0) {
                ifBLIN_QV1(r->flag) warn("setenv %s=%s", optarg, myspid);
            }
            break;
        case 'a':
            r->absq = strtol(optarg, NULL, 0);
            r->flag |= POPUSH_ABQ;
            break;
        case 'd': r->flag |= POPUSH_DMN;             break;
        case 'f': r->cnfname = optarg;               break;
        case 'p': r->pidname = optarg;               break;
        case 'q': r->flag &= ~BLIN_VERMASK;          break;
        case 'v': BLIN_VERBOSE(r->flag);             break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (~r->flag & POPUSH_DMN && daemon(0, 0)) {
        mesg(r->flag | POPUSH_CRI, DE("daemon"));
        exit(EX_UNAVAILABLE);
    }
    openlog("popushd", LOG_PID, LOG_DAEMON);
    if  ((r->kq = kqueue()) < 0) {
        mesg(r->flag | POPUSH_CRI, DE("kqueue"));
        ex = EX_OSERR;
        goto out;
    }
    state = F_STRT;
    if  (signal(SIGHUP   , sighup) == SIG_ERR) ifBLIN_QV1(r->flag) warn("SIGHUP");
    if  (signal(SIGINT   , sigint) == SIG_ERR) ifBLIN_QV1(r->flag) warn("SIGINT");
    if  (signal(SIGTERM  , sigint) == SIG_ERR) ifBLIN_QV1(r->flag) warn("SIGTERM");
    if  (signal(SIGXCPU  , sigint) == SIG_ERR) ifBLIN_QV1(r->flag) warn("SIGXCPU");
    if  (signal(SIGXFSZ  , sigint) == SIG_ERR) ifBLIN_QV1(r->flag) warn("SIGXFSZ");
    if  (signal(SIGVTALRM, sigint) == SIG_ERR) ifBLIN_QV1(r->flag) warn("SIGVTALRM");
    while (1) {
        ifBLIN_QV2(r->flag) fprintf(stderr, "Loop state %08X\n", state);
        if  ((state & F_STOP) && (unlink(r->pidname) < 0)) {
            mesg(r->flag, DE("PID file delete %s failed"), r->pidname);
        }
        if  (state & (F_STRT | F_SIGN)) {
            if  ((r->pidfile = fopen(r->pidname, "w")) == NULL) {
                mesg(r->flag, DE("PID file open %s failed"), r->pidname);
            } else {
                if (fprintf(r->pidfile, "%d\n", mypid) < 0) {
                    mesg(r->flag, DE("Can't write PID"));
                }
                if  (fclose(r->pidfile)) {
                    mesg(r->flag, DE("PID file not closed"));
        }   }   }
        r->pidfile = NULL;
        if  (state & (F_STRT | F_SIGN | F_SELF)) {
            if  (!(e = syntan(r))) {
                /* mesg(r->flag | POPUSH_ERX, DE("Syntax error")); */
            } else if (diffmerge(r, e)) {
                /* mesg(r->flag | POPUSH_ERX, DE("Configuration error")); */
            } else {
                cnfree(r->cnf);
                r->cnf = e;
                free(r->actionbit);
                r->actionbit = NULL;
        }   }
        if  (!r->cnf) {
            mesg(r->flag | POPUSH_ERX, DE("No config %s"), r->cnfname);
            ex = EX_TEMPFAIL;
            goto out;
        }
        if  (!r->actionbit && !(r->actionbit = createbit(MULAR_NEXT(r->cnf->cluster)))) {
            mesg(r->flag, DE("createbit"));
            ex = EX_TEMPFAIL;
            goto out;
        }
        if  (state & F_STRT) setact(r, r->cnf->strt);
        if  (state & F_STOP) setact(r, r->cnf->stop);
        if  (state & F_SIGN) setact(r, r->cnf->sign);
        if  (state & F_SELF) setaction(r, r->cnf->self);
        state &= F_STOP;
        if  ((ex = setevent(r))) {
            mesg(r->flag | POPUSH_ERX, DE("setevent"));
            goto out;
        }
        execevent(r);
        if  (state & F_STOP) {
            mesg(r->flag | POPUSH_CRI | POPUSH_ERX, DE("killed"));
            goto out;
        }
        if  (state) continue;
        bzero(&evnt, sizeof(evnt));
        c = kevent(r->kq, NULL, 0, &evnt, 1, NULL);
        if  (c > 0) setkact(r, &evnt);
    }
out:if  (!(state & F_STOP)) {
        /* XXXX restart */
    }
    closelog();
    exit(ex);
}
